@@ -1,12 +1,12 @@
-#ident "$Id: setup.c,v 1.29 2003/01/30 20:08:26 hpa Exp $"
+#ident "$Id: setup.c,v 1.47 2004/12/29 21:33:32 hpa Exp $"
 /* ----------------------------------------------------------------------- *
  *   
- *   Copyright 2001-2002 H. Peter Anvin - All Rights Reserved
+ *   Copyright 2001-2004 H. Peter Anvin - All Rights Reserved
  *
  *   This program is free software; you can redistribute it and/or modify
  *   it under the terms of the GNU General Public License as published by
  *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
- *   Bostom MA 02111-1307, USA; either version 2 of the License, or
+ *   Boston MA 02111-1307, USA; either version 2 of the License, or
  *   (at your option) any later version; incorporated herein by reference.
  *
  * ----------------------------------------------------------------------- */
@@ -15,8 +15,8 @@
 #include "e820.h"
 #include "conio.h"
 #include "version.h"
+#include "memdisk.h"
 
-/* A pointer to this is stored in the header */
 const char memdisk_version[] =
 "MEMDISK " VERSION " " DATE;
 const char copyright[] = 
@@ -62,25 +62,32 @@ typedef union {
 } dpt_t;
 
 struct patch_area {
+  uint32_t diskbuf;
+  uint32_t disksize;
+  uint16_t cmdline_off, cmdline_seg;
+  
+  uint32_t oldint13;
+  uint32_t oldint15;
+
+  uint16_t olddosmem;
+  uint8_t  bootloaderid;
+
+  uint8_t  _pad[3];
+  uint16_t memint1588;
+
   uint16_t cylinders;
   uint16_t heads;
   uint32_t sectors;
-  uint32_t disksize;
-  uint32_t diskbuf;
 
   uint32_t mem1mb;
   uint32_t mem16mb;
 
-  uint32_t oldint13;
-  uint32_t oldint15;
-
-  uint16_t memint1588;
-  uint16_t olddosmem;
-
   uint8_t  driveno;
   uint8_t  drivetype;
   uint8_t  drivecnt;
-  uint8_t  _pad1;
+  uint8_t  configflags;
+#define CONFIG_READONLY	0x01
+#define CONFIG_RAW	0x02
 
   uint16_t mystack;
   uint16_t statusptr;
@@ -116,112 +123,40 @@ struct setup_header {
   uint32_t initrd_addr_max;
 };
 
-const struct setup_header * const shdr = (struct setup_header *)0;
+const struct setup_header * const shdr = (struct setup_header *)(LOW_SEG << 4);
 
 /* Access to high memory */
 
-struct high_mover {
-  uint32_t resv1[4];		/* For the BIOS */
-  uint16_t src_limit;		/* 0xffff */
-  uint16_t src01;		/* Bytes 0-1 of src */
-  uint8_t src2;			/* Byte 2 of src */
-  uint8_t src_perms;		/* 0x93 */
-  uint8_t src_xperms;		/* 0x00 */
-  uint8_t src3;			/* Byte 3 of src */
-  uint16_t dst_limit;		/* 0xffff */
-  uint16_t dst01;		/* Bytes 0-1 of dst */
-  uint8_t dst2;			/* Byte 2 of dst */
-  uint8_t dst_perms;		/* 0x93 */
-  uint8_t dst_xperms;		/* 0x00 */
-  uint8_t dst3;			/* Byte 3 of dst */
-  uint32_t resv2[4];		/* For the BIOS */
-};
-
-/* Note: this version of high_bcopy() is limited to 64K */
-static void high_bcopy(uint32_t dst, uint32_t src, uint16_t len)
-{
-  static struct high_mover high_mover = 
-  {
-    { 0, 0, 0, 0 },
-    0xffff, 0, 0, 0x93, 0x00, 0,
-    0xffff, 0, 0, 0x93, 0x00, 0,
-    { 0, 0, 0, 0 }
-  };
-  
-  high_mover.src01 = (uint16_t)src;
-  high_mover.src2  = src >> 16;
-  high_mover.src3  = src >> 24;
-
-  high_mover.dst01 = (uint16_t)dst;
-  high_mover.dst2  = dst >> 16;
-  high_mover.dst3  = dst >> 24;
-
-  asm volatile("pushal ; "
-	       "pushfl ; "
-	       "movb $0x87,%%ah ; "
-	       "int $0x15 ; "
-	       "popfl ; "
-	       "popal"
-	       :: "S" (&high_mover), "c" (len >> 1)
-	       : "memory");
-}
-
-#define LOWSEG 0x0800		/* Should match init.S16 */
-
-static inline uint32_t
-ptr2linear(void *ptr)
-{
-  return (LOWSEG << 4) + (uint32_t)ptr;
-}
-
-static inline void
-copy_to_high(uint32_t dst, void *src, uint16_t len)
-{
-  high_bcopy(dst, ptr2linear(src), len);
-}
-
-static inline void
-copy_from_high(void *dst, uint32_t src, uint16_t len)
-{
-  high_bcopy(ptr2linear(dst), src, len);
-}
-
 /* Access to objects in the zero page */
 static inline void
 wrz_8(uint32_t addr, uint8_t data)
 {
-  asm volatile("movb %0,%%fs:%1" :: "ri" (data), "m" (*(uint8_t *)addr));
+  *((uint8_t *)addr) = data;
 }
 static inline void
 wrz_16(uint32_t addr, uint16_t data)
 {
-  asm volatile("movw %0,%%fs:%1" :: "ri" (data), "m" (*(uint16_t *)addr));
+  *((uint16_t *)addr) = data;
 }
 static inline void
 wrz_32(uint32_t addr, uint32_t data)
 {
-  asm volatile("movl %0,%%fs:%1" :: "ri" (data), "m" (*(uint32_t *)addr));
+  *((uint32_t *)addr) = data;
 }
 static inline uint8_t
 rdz_8(uint32_t addr)
 {
-  uint8_t data;
-  asm volatile("movb %%fs:%1,%0" : "=r" (data) : "m" (*(uint8_t *)addr));
-  return data;
+  return *((uint8_t *)addr);
 }
 static inline uint16_t
 rdz_16(uint32_t addr)
 {
-  uint16_t data;
-  asm volatile("movw %%fs:%1,%0" : "=r" (data) : "m" (*(uint16_t *)addr));
-  return data;
+  return *((uint16_t *)addr);
 }
 static inline uint32_t
 rdz_32(uint32_t addr)
 {
-  uint32_t data;
-  asm volatile("movl %%fs:%1,%0" : "=r" (data) : "m" (*(uint32_t *)addr));
-  return data;
+  return *((uint32_t *)addr);
 }
 
 /* Addresses in the zero page */
@@ -290,6 +225,111 @@ const char *getcmditem(const char *what)
   return CMD_NOTFOUND;
 }
 
+/*
+ * Check to see if this is a gzip image
+ */
+#define UNZIP_ALIGN 512
+
+extern void _end;		/* Symbol signalling end of data */
+
+void unzip_if_needed(uint32_t *where_p, uint32_t *size_p)
+{
+  uint32_t where = *where_p;
+  uint32_t size = *size_p;
+  uint32_t zbytes;
+  uint32_t startrange, endrange;
+  uint32_t gzdatasize, gzwhere;
+  uint32_t orig_crc, offset;
+  uint32_t target = 0;
+  int i, okmem;
+
+  /* Is it a gzip image? */
+  if (check_zip ((void *)where, size, &zbytes, &gzdatasize,
+                 &orig_crc, &offset) == 0) {
+
+    if (offset + zbytes > size) {
+      /* Assertion failure; check_zip is supposed to guarantee this
+         never happens. */
+      puts("internal error: check_zip returned nonsense\n");
+      die();
+    }
+
+    /* Find a good place to put it: search memory ranges in descending order
+       until we find one that is legal and fits */
+    okmem = 0;
+    for ( i = nranges-1 ; i >= 0 ; i-- ) {
+      /* We can't use > 4G memory (32 bits only.)  Truncate to 2^32-1
+	 so we don't have to deal with funny wraparound issues. */
+      
+      /* Must be memory */
+      if ( ranges[i].type != 1 )
+	continue;
+
+      /* Range start */
+      if ( ranges[i].start >= 0xFFFFFFFF )
+	continue;
+      startrange = (uint32_t)ranges[i].start;
+
+      /* Range end (0 for end means 2^64) */
+      endrange = ((ranges[i+1].start >= 0xFFFFFFFF ||
+		   ranges[i+1].start == 0)
+		  ? 0xFFFFFFFF : (uint32_t)ranges[i+1].start);
+
+      /* Make sure we don't overwrite ourselves */
+      if ( startrange < (uint32_t)&_end )
+	startrange = (uint32_t)&_end;
+
+      /* Allow for alignment */
+      startrange = (ranges[i].start + (UNZIP_ALIGN-1)) & ~(UNZIP_ALIGN-1);
+
+      /* In case we just killed the whole range... */
+      if ( startrange >= endrange )
+	continue;
+
+      /* Must be large enough... don't rely on gzwhere for this (wraparound) */
+      if ( endrange-startrange < gzdatasize )
+	continue;
+
+      /* This is where the gz image should be put if we put it in this range */
+      gzwhere = (endrange - gzdatasize) & ~(UNZIP_ALIGN-1);
+
+      /* Cast to uint64_t just in case we're flush with the top byte */
+      if ( (uint64_t)where+size >= gzwhere && where < endrange ) {
+	/* Need to move source data to avoid compressed/uncompressed overlap */
+	uint32_t newwhere;
+	
+	if ( gzwhere-startrange < size )
+	  continue;		/* Can't fit both old and new */
+	
+	newwhere = (gzwhere - size) & ~(UNZIP_ALIGN-1);
+	printf("Moving compressed data from 0x%08x to 0x%08x\n",
+	       where, newwhere);
+	
+	/* Our memcpy() is OK, because we always move from a higher
+	   address to a lower one */
+	memcpy((void *)newwhere, (void *)where, size);
+	where = newwhere;
+      }
+
+      target = gzwhere;
+      okmem = 1;
+      break;
+    }
+
+    if ( !okmem ) {
+      puts("Not enough memory to decompress image\n");
+      die();
+    }
+
+    printf("gzip image: decompressed addr 0x%08x, len 0x%08x: ",
+	   target, gzdatasize);
+
+    *size_p  = gzdatasize;
+    *where_p = (uint32_t)unzip((void *)(where + offset), zbytes,
+                               gzdatasize, orig_crc, (void *)target);
+  }
+}
+
 /*
  * Figure out the "geometry" of the disk in question
  */
@@ -303,11 +343,14 @@ struct geometry {
 
 static const struct geometry geometries[] =
 { 
-  {  720, 40,  2,  9, 0, 0x01, 0 }, /*  360 K */
-  { 1440, 80,  2,  9, 0, 0x03, 0 }, /*  720 K*/
-  { 2400, 80,  2, 15, 0, 0x02, 0 }, /* 1200 K */
-  { 2880, 80,  2, 18, 0, 0x04, 0 }, /* 1440 K */
-  { 5760, 80,  2, 36, 0, 0x06, 0 }, /* 2880 K */
+  {  360*2, 40,  2,  9, 0, 0x01, 0 }, /*  360 K */
+  {  720*2, 80,  2,  9, 0, 0x03, 0 }, /*  720 K*/
+  { 1200*2, 80,  2, 15, 0, 0x02, 0 }, /* 1200 K */
+  { 1440*2, 80,  2, 18, 0, 0x04, 0 }, /* 1440 K */
+  { 1680*2, 80,  2, 21, 0, 0x04, 0 }, /* 1680 K */
+  { 1722*2, 82,  2, 21, 0, 0x04, 0 }, /* 1722 K */
+  { 2880*2, 80,  2, 36, 0, 0x06, 0 }, /* 2880 K */
+  { 3840*2, 80,  2, 48, 0, 0x06, 0 }, /* 3840 K */
 };
 #define known_geometries (sizeof(geometries)/sizeof(struct geometry))
 
@@ -331,6 +374,8 @@ struct dosemu_header {
   uint8_t pad[105];
 } __attribute__((packed));
 
+#define FOUR(a,b,c,d) (((a) << 24)|((b) << 16)|((c) << 8)|(d))
+
 const struct geometry *get_disk_image_geometry(uint32_t where, uint32_t size)
 {
   static struct geometry hd_geometry = { 0, 0, 0, 0, 0, 0, 0x80 };
@@ -361,8 +406,9 @@ const struct geometry *get_disk_image_geometry(uint32_t where, uint32_t size)
   hd_geometry.offset  = offset;
 
   /* Do we have a DOSEMU header? */
-  copy_from_high(&dosemu, where+hd_geometry.offset, sizeof dosemu);
-  if ( !__builtin_memcmp("DOSEMU", dosemu.magic, 7) ) {
+  memcpy(&dosemu, (char *)where+hd_geometry.offset, sizeof dosemu);
+  if ( ((unsigned long *)dosemu.magic)[0] == FOUR('D','O','S','E') &&
+       (((unsigned long *)dosemu.magic)[1] & 0xffffff) == FOUR('M','U',0,0) ) {
     /* Always a hard disk unless overruled by command-line options */
     hd_geometry.driveno = 0x80;
     hd_geometry.type = 0;
@@ -394,7 +440,7 @@ const struct geometry *get_disk_image_geometry(uint32_t where, uint32_t size)
   if ( (hd_geometry.c == 0) || (hd_geometry.h == 0) ||
        (hd_geometry.s == 0) ) {
     /* Hard disk image, need to examine the partition table for geometry */
-    copy_from_high(&ptab, where+hd_geometry.offset+(512-2-4*16), sizeof ptab);
+    memcpy(&ptab, (char *)where+hd_geometry.offset+(512-2-4*16), sizeof ptab);
     
     max_c = max_h = 0;  max_s = 1;
     for ( i = 0 ; i < 4 ; i++ ) {
@@ -450,14 +496,17 @@ void __attribute__((noreturn)) die(void)
     asm volatile("hlt");
 }
 
-#define STACK_NEEDED	128	/* Number of bytes of stack */
+#define STACK_NEEDED	512	/* Number of bytes of stack */
 
 /*
  * Actual setup routine
  * Returns the drive number (which is then passed in %dl to the
  * called routine.)
  */
-uint32_t setup(void)
+syscall_t syscall;
+void *sys_bounce;
+
+uint32_t setup(syscall_t cs_syscall, void *cs_bounce)
 {
   unsigned int bin_size = (int) &_binary_memdisk_bin_size;
   struct memdisk_header *hptr;
@@ -466,10 +515,14 @@ uint32_t setup(void)
   uint32_t driverptr, driveraddr;
   uint16_t dosmem_k;
   uint32_t stddosmem;
-  uint8_t status;
-  uint16_t exitcode;
   const struct geometry *geometry;
-  int total_size;
+  int total_size, cmdlinelen;
+  com32sys_t regs;
+  uint32_t ramdisk_image, ramdisk_size;
+
+  /* Set up global variables */
+  syscall = cs_syscall;
+  sys_bounce = cs_bounce;
 
   /* Show signs of life */
   printf("%s  %s\n", memdisk_version, copyright);
@@ -479,29 +532,27 @@ uint32_t setup(void)
     die();
   }
 
-  printf("Ramdisk at 0x%08x, length 0x%08x\n",
-	 shdr->ramdisk_image, shdr->ramdisk_size);
+  ramdisk_image = shdr->ramdisk_image;
+  ramdisk_size  = shdr->ramdisk_size;
 
-  geometry = get_disk_image_geometry(shdr->ramdisk_image, shdr->ramdisk_size);
+  e820map_init();		/* Initialize memory data structure */
+  get_mem();			/* Query BIOS for memory map */
+  parse_mem();			/* Parse memory map */
+
+  printf("Ramdisk at 0x%08x, length 0x%08x\n",
+	 ramdisk_image, ramdisk_size);
+
+  unzip_if_needed(&ramdisk_image, &ramdisk_size);
+
+  geometry = get_disk_image_geometry(ramdisk_image, ramdisk_size);
 
   printf("Disk is %s, %u K, C/H/S = %u/%u/%u\n",
 	 geometry->driveno ? "hard disk" : "floppy",
 	 geometry->sectors >> 1,
 	 geometry->c, geometry->h, geometry->s);
 
-  e820map_init();		/* Initialize memory data structure */
-  get_mem();			/* Query BIOS for memory map */
-  parse_mem();			/* Parse memory map */
-
-  printf("dos_mem  = %#10x (%u K)\n"
-	 "low_mem  = %#10x (%u K)\n"
-	 "high_mem = %#10x (%u K)\n",
-	 dos_mem, dos_mem >> 10,
-	 low_mem, low_mem >> 10,
-	 high_mem, high_mem >> 10);
-
   /* Reserve the ramdisk memory */
-  insertrange(shdr->ramdisk_image, shdr->ramdisk_size, 2);
+  insertrange(ramdisk_image, ramdisk_size, 2);
   parse_mem();			/* Recompute variables */
 
   /* Figure out where it needs to go */
@@ -521,9 +572,22 @@ uint32_t setup(void)
   pptr->heads     = geometry->h;
   pptr->sectors   = geometry->s;
   pptr->disksize  = geometry->sectors;
-  pptr->diskbuf   = shdr->ramdisk_image + geometry->offset;
+  pptr->diskbuf   = ramdisk_image + geometry->offset;
   pptr->statusptr = (geometry->driveno & 0x80) ? 0x474 : 0x441;
 
+  pptr->bootloaderid = shdr->type_of_loader;
+
+  pptr->configflags = 0;
+  /* Set config flags */
+  if ( getcmditem("ro") != CMD_NOTFOUND ) {
+    puts("Marking disk readonly\n");
+    pptr->configflags |= CONFIG_READONLY;
+  }
+  if ( getcmditem("raw") != CMD_NOTFOUND ) {
+    puts("Using raw access to high memory\n");
+    pptr->configflags |= CONFIG_RAW;
+  }
+
   /* Set up a drive parameter table */
   if ( geometry->driveno & 0x80 ) {
     /* Hard disk */
@@ -548,13 +612,17 @@ uint32_t setup(void)
     pptr->dpt.fd.old_fd_dpt = rdz_32(BIOS_INT1E);
   }
 
-  /* The size is given by hptr->total_size plus the size of the
-     E820 map -- 12 bytes per range; we may need as many as
-     2 additional ranges plus the terminating range, over what
-     nranges currently show. */
-
-  total_size = hptr->total_size + (nranges+3)*12 + STACK_NEEDED;
-  printf("Total size needed = %u bytes\n", total_size);
+  /* The size is given by hptr->total_size plus the size of the E820
+     map -- 12 bytes per range; we may need as many as 2 additional
+     ranges (each insertrange() can worst-case turn 1 area into 3)
+     plus the terminating range, over what nranges currently show. */
+  cmdlinelen = strlen(shdr->cmdline)+1;
+  total_size  =  hptr->total_size; 		/* Actual memdisk code */
+  total_size += (nranges+3)*sizeof(ranges[0]);  /* E820 memory ranges */
+  total_size += cmdlinelen;	                /* Command line */
+  total_size += STACK_NEEDED;	                /* Stack */
+  printf("Total size needed = %u bytes, allocating %uK\n",
+	 total_size, (total_size+0x3ff) >> 10);
 
   if ( total_size > dos_mem ) {
     puts("MEMDISK: Insufficient low memory\n");
@@ -572,13 +640,6 @@ uint32_t setup(void)
   insertrange(driveraddr, dos_mem-driveraddr, 2);
   parse_mem();
 
-  printf("dos_mem  = %#10x (%u K)\n"
-	 "low_mem  = %#10x (%u K)\n"
-	 "high_mem = %#10x (%u K)\n",
-	 dos_mem, dos_mem >> 10,
-	 low_mem, low_mem >> 10,
-	 high_mem, high_mem >> 10);
-
   pptr->mem1mb     = low_mem  >> 10;
   pptr->mem16mb    = high_mem >> 16;
   if ( low_mem == (15 << 20) ) {
@@ -610,62 +671,42 @@ uint32_t setup(void)
   }
 
   /* Query drive parameters of this type */
-  {
-    uint16_t bpt_es, bpt_di;
-    uint8_t cf, dl;
-
-    asm volatile("pushw %%es ; "
-		 "xorw %1,%1 ; "
-		 "movw %1,%%es ; "
-		 "movb $0x08,%%ah ; "
-		 "int $0x13 ; "
-		 "setc %2 ; "
-		 "movw %%es,%0 ;"
-		 "popw %%es"
-		 : "=a" (bpt_es), "=D" (bpt_di),
-		 "=c" (cf), "=d" (dl)
-		 : "d" (geometry->driveno & 0x80)
-		 : "esi", "ebx", "ebp");
-
-    if ( cf ) {
-      printf("INT 13 08: Failure, assuming this is the only drive\n");
-      pptr->drivecnt = 1;
-    } else {
-      printf("INT 13 08: Success, count = %u, BPT = %04x:%04x\n",
-	     dl, bpt_es, bpt_di);
-      pptr->drivecnt = dl+1;
-    }
+  memset(&regs, 0, sizeof regs);
+  regs.es = 0;
+  regs.eax.b[1] = 0x08;
+  regs.edx.b[0] = geometry->driveno;
+  syscall(0x13, &regs, &regs);
+  
+  if ( regs.eflags.l & 1 ) {
+    printf("INT 13 08: Failure, assuming this is the only drive\n");
+    pptr->drivecnt = 1;
+  } else {
+    printf("INT 13 08: Success, count = %u, BPT = %04x:%04x\n",
+	   regs.edx.b[0], regs.es, regs.edi.w[0]);
+    pptr->drivecnt = regs.edx.b[0]+1;
   }
 
-  /* Copy driver followed by E820 table */
-  asm volatile("pushal ; "
-	       "pushw %%es ; "
-	       "movw %0,%%es ; "
-	       "cld ; "
-	       "rep ; movsl %%ds:(%%si), %%es:(%%di) ; "
-	       "movw %1,%%cx ; "
-	       "movw %2,%%si ; "
-	       "rep ; movsl %%ds:(%%si), %%es:(%%di) ; "
-	       "popw %%es ; "
-	       "popal"
-	       :: "r" (driverseg),
-	       "r" ((uint16_t)((nranges+1)*3)), /* 3 dwords/range */
-	       "r" ((uint16_t)&ranges),
-	       "c" (bin_size >> 2),
-	       "S" (&_binary_memdisk_bin_start),
-	       "D" (0));
+  /* Pointer to the command line */
+  pptr->cmdline_off = bin_size + (nranges+1)*sizeof(ranges[0]);
+  pptr->cmdline_seg = driverseg;
+
+  /* Copy driver followed by E820 table followed by command line */
+  {
+    unsigned char *dpp = (unsigned char *)(driverseg << 4);
+    dpp = memcpy_endptr(dpp, &_binary_memdisk_bin_start, bin_size);
+    dpp = memcpy_endptr(dpp, ranges, (nranges+1)*sizeof(ranges[0]));
+    dpp = memcpy_endptr(dpp, shdr->cmdline, cmdlinelen+1);
+  }
 
   /* Install the interrupt handlers */
-  {
-    printf("old: int13 = %08x  int15 = %08x\n",
-	   rdz_32(BIOS_INT13), rdz_32(BIOS_INT15));
-
-    wrz_32(BIOS_INT13, driverptr+hptr->int13_offs);
-    wrz_32(BIOS_INT15, driverptr+hptr->int15_offs);
-
-    printf("new: int13 = %08x  int15 = %08x\n",
-	   rdz_32(BIOS_INT13), rdz_32(BIOS_INT15));
-  }
+  printf("old: int13 = %08x  int15 = %08x\n",
+	 rdz_32(BIOS_INT13), rdz_32(BIOS_INT15));
+  
+  wrz_32(BIOS_INT13, driverptr+hptr->int13_offs);
+  wrz_32(BIOS_INT15, driverptr+hptr->int15_offs);
+  
+  printf("new: int13 = %08x  int15 = %08x\n",
+	 rdz_32(BIOS_INT13), rdz_32(BIOS_INT15));
 
   /* Update various BIOS magic data areas (gotta love this shit) */
 
@@ -673,7 +714,6 @@ uint32_t setup(void)
     /* Update BIOS hard disk count */
     wrz_8(BIOS_HD_COUNT, rdz_8(BIOS_HD_COUNT)+1);
   } else {
-#if 1				/* Apparently this is NOT wanted... */
     /* Update BIOS floppy disk count */
     uint8_t equip = rdz_8(BIOS_EQUIP);
     if ( equip & 1 ) {
@@ -685,41 +725,31 @@ uint32_t setup(void)
       equip &= ~(3 << 6);
     }
     wrz_8(BIOS_EQUIP, equip);
-#endif
   }
 
-  /* Reboot into the new "disk" */
-  asm volatile("pushl %%ebp ; "
-	       "pushl %%edx ; "
-	       "pushw %%es ; "
-	       "xorw %%cx,%%cx ; "
-	       "movw %%cx,%%es ; "
-	       "incw %%cx ; "
-	       "movw $0x0201,%%ax ; "
-	       "movw $0x7c00,%%bx ; "
-	       "int $0x13 ; "
-	       "popw %%es ; "
-	       "popl %%edx ; "
-	       "popl %%ebp ; "
-	       "setc %0"
-	       : "=rm" (status), "=a" (exitcode)
-	       : "d" ((uint16_t)geometry->driveno)
-	       : "ecx", "ebx", "esi", "edi");
+  /* Reboot into the new "disk"; this is also a test for the interrupt hooks */
+  puts("Loading boot sector... ");
 
-  if ( status ) {
+  memset(&regs, 0, sizeof regs);
+  // regs.es = 0;
+  regs.eax.w[0] = 0x0201;	/* Read sector */
+  regs.ebx.w[0] = 0x7c00;	/* 0000:7C00 */
+  regs.ecx.w[0] = 1;		/* One sector */
+  regs.edx.w[0] = geometry->driveno;
+  syscall(0x13, &regs, &regs);
+
+  if ( regs.eflags.l & 1 ) {
     puts("MEMDISK: Failed to load new boot sector\n");
     die();
   }
   
   if ( getcmditem("pause") != CMD_NOTFOUND ) {
-    puts("Press any key to boot... ");
-    asm volatile("pushal ; "
-		 "xorw %ax,%ax ; "
-		 "int $0x16 ; "
-		 "popal");
+    puts("press any key to boot... ");
+    regs.eax.w[0] = 0;
+    syscall(0x16, &regs, NULL);
   }
 
-  puts("Booting...\n");
+  puts("booting...\n");
 
   /* On return the assembly code will jump to the boot vector */
   return geometry->driveno;
